import asyncio
import subprocess
import tempfile

from jcloud.flow import CloudFlow


def deploy_wolf(path: str, name: str, env_file: str = None):
    return CloudFlow(path=path, name=name, env_file=env_file).__enter__()


def terminate_wolf(flow_id: str):
    CloudFlow(flow_id=flow_id).__exit__()


def status_wolf(flow_id):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(CloudFlow(flow_id=flow_id).status)


def list_all_wolf(status='READY', namespace='nowapi'):
    loop = asyncio.get_event_loop()
    flows = loop.run_until_complete(CloudFlow().list_all(status=status))
    if flows is None:
        return []
    # filter by namespace - if the namespace is contained in the flow name
    flows = [flow for flow in flows if namespace in flow['name']]
    return flows


def cmd(command, std_output=False, wait=True):
    if isinstance(command, str):
        command = command.split()
    if not std_output:
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    else:
        process = subprocess.Popen(command)
    if wait:
        output, error = process.communicate()
        return output, error


def which(executable: str) -> bool:
    return bool(cmd('which ' + executable)[0])


def apply_replace(f_in, replace_dict, kubectl_path):
    with open(f_in, "r") as fin:
        with tempfile.NamedTemporaryFile(mode='w') as fout:
            for line in fin.readlines():
                for key, val in replace_dict.items():
                    line = line.replace('{' + key + '}', str(val))
                fout.write(line)
            fout.flush()
            cmd(f'{kubectl_path} apply -f {fout.name}')



import os.path
import pickle
from collections import defaultdict
from copy import deepcopy
from sys import maxsize
from typing import Dict, List, Optional

import annlite
from jina import Document, DocumentArray
from jina.logging.logger import JinaLogger
from now_executors.NOWAuthExecutor.executor import NOWAuthExecutor as Executor
from now_executors.NOWAuthExecutor.executor import SecurityLevel, secure_request


class NOWAnnLiteIndexer(Executor):
    """
    A simple Indexer based on PQLite that stores all the Document data together in a local LMDB store.

    To be used as a hybrid indexer, supporting pre-filtering searching.
    """

    def __init__(
        self,
        dim: int = 0,
        metric: str = 'cosine',
        limit: int = 10,
        ef_construction: int = 200,
        ef_query: int = 50,
        max_connection: int = 16,
        include_metadata: bool = True,
        index_traversal_paths: str = '@r',
        search_traversal_paths: str = '@r',
        columns: Optional[List] = None,
        serialize_config: Optional[Dict] = None,
        *args,
        **kwargs,
    ):
        """
        :param dim: Dimensionality of vectors to index
        :param metric: Distance metric type. Can be 'euclidean', 'inner_product', or 'cosine'
        :param include_metadata: If True, return the document metadata in response
        :param limit: Number of results to get for each query document in search
        :param ef_construction: The construction time/accuracy trade-off
        :param ef_query: The query time accuracy/speed trade-off
        :param max_connection: The maximum number of outgoing connections in the
            graph (the "M" parameter)
        :param index_traversal_paths: Default traversal paths on docs
                (used for indexing, delete and update), e.g. '@r', '@c', '@r,c'
        :param search_traversal_paths: Default traversal paths on docs
        (used for search), e.g. '@r', '@c', '@r,c'
        :param columns: List of tuples of the form (column_name, str_type). Here str_type must be a string that can be
                parsed as a valid Python type.
        :param serialize_config: The configurations used for serializing documents, e.g., {'protocol': 'pickle'}
        """
        super().__init__(*args, **kwargs)
        self.logger = JinaLogger(self.__class__.__name__)

        assert dim > 0, 'Please specify the dimension of the vectors to index!'

        self.metric = metric
        self.limit = limit
        self.include_metadata = include_metadata
        self.index_traversal_paths = index_traversal_paths
        self.search_traversal_paths = search_traversal_paths
        self._valid_input_columns = ['str', 'float', 'int']

        if columns:
            corrected_list = []
            for i in range(0, len(columns), 2):
                corrected_list.append((columns[i], columns[i + 1]))
            columns = corrected_list
            cols = []
            for n, t in columns:
                assert (
                    t in self._valid_input_columns
                ), f'column of type={t} is not supported. Supported types are {self._valid_input_columns}'
                cols.append((n, eval(t)))
            columns = cols

        self._index = annlite.AnnLite(
            n_dim=dim,
            metric=metric,
            columns=columns,
            ef_construction=ef_construction,
            ef_query=ef_query,
            max_connection=max_connection,
            data_path=self.workspace or './workspace',
            serialize_config=serialize_config or {},
            **kwargs,
        )
        self.da = DocumentArray()
        for cell_id in range(self._index.n_cells):
            for docs in self._index.documents_generator(cell_id, batch_size=10240):
                self.extend_inmemory_docs_and_tags(docs)

        self.da = DocumentArray(sorted([d for d in self.da], key=lambda x: x.id))

        if os.path.isfile(os.path.join(self.workspace, 'tags.pkl')):
            with open(os.path.join(self.workspace, 'tags.pkl'), 'rb') as f:
                self.doc_id_tags = pickle.load(f)
        else:
            self.doc_id_tags = defaultdict(dict)

    def extend_inmemory_docs_and_tags(self, docs):
        """Extend the in-memory DocumentArray with new documents"""
        for d in docs:
            self.da.append(
                Document(id=d.id, uri=d.uri, tags=d.tags, parent_id=d.parent_id)
            )
            self.doc_id_tags[d.id] = d.tags
        self.save_tags()

    def update_inmemory_docs_and_tags(self, docs):
        """Update documents in the in-memory DocumentArray"""
        for d in docs:
            self.da[d.id] = d
            self.doc_id_tags[d.id] = d.tags
        self.save_tags()

    def delete_inmemory_docs_and_tags(self, docs):
        """Delete documents from the in-memory DocumentArray"""
        for d in docs:
            del self.da[d.id]
            self.doc_id_tags.pop(d.id, None)
        self.save_tags()

    @secure_request(on='/index', level=SecurityLevel.USER)
    def index(
        self, docs: Optional[DocumentArray] = None, parameters: dict = {}, **kwargs
    ):
        """Index new documents

        :param docs: the Documents to index
        :param parameters: dictionary with options for indexing
        Keys accepted:
            - 'traversal_paths' (str): traversal path for the docs
        """
        if not docs:
            return

        traversal_paths = parameters.get('traversal_paths', self.index_traversal_paths)
        flat_docs = docs[traversal_paths]
        if len(flat_docs) == 0:
            return

        self._index.index(flat_docs)
        self.extend_inmemory_docs_and_tags(flat_docs)

        return DocumentArray([])

    @secure_request(on='/tags', level=SecurityLevel.USER)
    def get_tags_and_values(self, **kwargs):
        """Returns tags and their possible values

        for example if indexed docs are the following:
            docs = DocumentArray([
                Document(.., tags={'color':'red'}),
                Document(.., tags={'color':'blue'}),
                Document(.., tags={'greeting':'hello'}),
            ])

        the resulting response would be a document array with
        one document containg a dictionary in tags like the following:
        {'tags':{'color':['red', 'blue'], 'greeting':['hello']}}
        """

        count_dict = defaultdict(lambda: defaultdict(int))
        for tags in self.doc_id_tags.values():
            for key, value in tags.items():
                count_dict[key][value] += 1

        for key, inside_dict in count_dict.items():
            count_dict[key] = dict(
                sorted(inside_dict.items(), key=lambda item: item[1])
            )
            count_dict[key] = list((count_dict[key]))[-10:]

        return DocumentArray([Document(text='tags', tags={'tags': dict(count_dict)})])

    @secure_request(on='/update', level=SecurityLevel.USER)
    def update(
        self, docs: Optional[DocumentArray] = None, parameters: dict = {}, **kwargs
    ):
        """Update existing documents

        :param docs: the Documents to update
        :param parameters: dictionary with options for updating
        Keys accepted:

            - 'traversal_paths' (str): traversal path for the docs
        """

        if not docs:
            return

        traversal_paths = parameters.get('traversal_paths', self.index_traversal_paths)
        flat_docs = docs[traversal_paths]
        if len(flat_docs) == 0:
            return

        self.update_inmemory_docs_and_tags(flat_docs)
        self._index.update(flat_docs)

    @secure_request(on='/delete', level=SecurityLevel.USER)
    def delete(self, parameters: dict = {}, **kwargs):
        """
        Delete endpoint to delete document/documents from the index.
        Filter conditions can be passed to select documents for deletion.
        """
        filter = parameters.get("filter", {})
        if filter:
            filtered_docs = deepcopy(self.da.find(filter=filter))
            self.delete_inmemory_docs_and_tags(filtered_docs)
            self._index.delete(filtered_docs)

        return DocumentArray()

    @secure_request(on='/list', level=SecurityLevel.USER)
    def list(self, parameters: dict = {}, **kwargs):
        """List all indexed documents.
        :param parameters: dictionary with limit and offset
        - offset (int): number of documents to skip
        - limit (int): number of retrieved documents
        """
        limit = int(parameters.get('limit', maxsize))
        offset = int(parameters.get('offset', 0))
        # add removal of duplicates
        traversal_paths = parameters.get('traversal_paths', self.search_traversal_paths)
        if traversal_paths == '@c':
            docs = DocumentArray()
            chunks_size = int(parameters.get('chunks_size', 3))
            parent_ids = set()
            for d in self.da[offset * chunks_size :]:
                if len(parent_ids) == limit:
                    break
                if d.parent_id in parent_ids:
                    continue
                parent_ids.add(d.parent_id)
                docs.append(d)
        else:
            docs = self.da[offset : offset + limit]
        return docs

    @secure_request(on='/search', level=SecurityLevel.USER)
    def search(
        self, docs: Optional[DocumentArray] = None, parameters: dict = {}, **kwargs
    ):
        """Perform a vector similarity search and retrieve Document matches

        Search can be performed with candidate filtering. Filters are a triplet (column,operator,value).
        More than a filter can be applied during search. Therefore, conditions for a filter are specified as a list triplets.
        Each triplet contains:

        - column: Column used to filter.
        - operator: Binary operation between two values. Some supported operators include `['>','<','=','<=','>=']`.
        - value: value used to compare a candidate.

        :param docs: the Documents to search with
        :param parameters: dictionary for parameters for the search operation
        Keys accepted:

            - 'filter' (dict): the filtering conditions on document tags
            - 'traversal_paths' (str): traversal paths for the docs
            - 'limit' (int): nr of matches to get per Document
        """
        if not docs:
            return

        limit = int(parameters.get('limit', self.limit))
        search_filter = parameters.get('filter', {})
        include_metadata = bool(
            parameters.get('include_metadata', self.include_metadata)
        )

        traversal_paths = parameters.get('traversal_paths', self.search_traversal_paths)
        flat_docs = docs[traversal_paths]
        if len(flat_docs) == 0:
            return

        if self.search_traversal_paths == '@c':
            retrieval_limit = limit * 3
        else:
            retrieval_limit = limit

        self._index.search(
            flat_docs,
            filter=search_filter,
            limit=retrieval_limit,
            include_metadata=include_metadata,
        )

        if self.search_traversal_paths == '@c':
            docs = docs[0].chunks
            for d in docs:
                unique_matches = []
                parent_ids = set()
                d.embedding = None
                for m in d.matches:
                    m.embedding = None
                    if m.parent_id in parent_ids:
                        continue
                    unique_matches.append(m)
                    parent_ids.add(m.parent_id)
                    if len(unique_matches) == limit:
                        break
                d.matches = unique_matches
        return docs

    @secure_request(on='/status', level=SecurityLevel.USER)
    def status(self, **kwargs) -> DocumentArray:
        """Return the document containing status information about the indexer.

        The status will contain information on the total number of indexed and deleted
        documents, and on the number of (searchable) documents currently in the index.
        """

        status = Document(tags=self._index.stat)
        return DocumentArray([status])

    @secure_request(on='/clear', level=SecurityLevel.USER)
    def clear(self, **kwargs):
        """Clear the index of all entries."""
        self._index.clear()

    def close(self, **kwargs):
        """Close the index."""
        self._index.close()

    def save_tags(self):
        with open(os.path.join(self.workspace, 'tags.pkl'), 'wb') as f:
            pickle.dump(self.doc_id_tags, f)



import copy
import os.path
from collections import defaultdict
from copy import deepcopy

import numpy as np
from finetuner.tuner.evaluation import Evaluator

from now.hub.head_encoder.head_encoder import FineTunedLinearHeadEncoder
from now.utils import plot_metrics, save_before_after_image, visual_result

np.seterr(divide='ignore', invalid='ignore')

random_state = 42


def show_improvement(
    data,
    query,
    index,
    query_all,
    index_all,
    final_layer_output_dim,
    embedding_size,
    finetuned_model_path,
    top_x=10,
    class_label='finetuner_label',
):
    """
    The user should see some visualizations on how the pre-trained model performed compared to the fine-tuned to show
    the power of fine-tuner.
    1. show search results before/ after in a flashy way
    2. show the linear probing ability of the model before/after (kalims graphics)

    Ideas:
    a) you could start an HTML page
    b) you could plt.show() some graphics and save them on disc
    """
    # 1: Show the linear probing ability of the model before/after
    # Step 1: Prepare filtered dataset
    doc_to_classes = defaultdict(int)
    for d in index:
        doc_to_classes[d.tags[class_label]] += 1

    # sort the filtered doc by their count in decreasing order
    tuples = [(category, count) for category, count in doc_to_classes.items()]
    tuples.sort(key=lambda x: x[1], reverse=True)
    # pick top_x classes from the docs
    top_x_classes = [x[0] for idx, x in enumerate(tuples) if idx < top_x]

    # now filter the doc based on the top_x_classes and take max 2K documents from each classes
    labels = []
    subset = []
    label_dict = {}
    label_id = 0
    for d in index:
        cat = d.tags[class_label]
        if cat not in top_x_classes:
            continue
        if cat not in label_dict:
            label_dict[cat] = label_id
            label_id += 1
        if not doc_to_classes[cat] > 2000:
            subset.append(copy.deepcopy(d))
            labels.append(label_dict[cat])
    # filtered_doc = DocumentArray(subset)

    # Step 2: Get pretrained and finetuned embeddings
    # pretrained_embed = filtered_doc.embeddings
    mean_path = os.path.dirname(finetuned_model_path)
    mean_path = os.path.realpath(mean_path)
    finetuned_encoder = FineTunedLinearHeadEncoder(
        final_layer_output_dim,
        embedding_size,
        model_path=finetuned_model_path,
        mean_path=mean_path,
    )
    # finetuned_docs = finetuned_encoder.encode(filtered_doc)
    # finetuned_embed = finetuned_docs.embeddings

    # # Step 3: Calling SVM on finetuned & pretrained embed
    # get_pr_curve(pretrained_embed, labels, title='pretrained_pr')
    # get_pr_curve(finetuned_embed, labels, title='finetuned_pr')

    # 2. Get search results/metric and plot together
    # If all `finetuner_label` are unique then do not highlight with any color
    unique_labels = set([doc.tags[class_label] for doc in index])
    unique = len(unique_labels) == len(index)
    new_query = finetuned_encoder.encode(deepcopy(query))
    new_index = finetuned_encoder.encode(deepcopy(index))
    # Step 1: Pre-trained
    query.match(index, limit=9, exclude_self=True)
    visual_result(
        data, query, output='pretrained.png', label=class_label, unique=unique
    )
    evaluator = Evaluator(query_data=query_all, index_data=index_all)
    ev = evaluator.evaluate()
    plot_metrics(ev, 'pretrained_m.png')

    # Fine-tuned
    new_query.match(new_index, limit=9, exclude_self=True)
    visual_result(
        data, new_query, output='finetuned.png', label=class_label, unique=unique
    )
    new_query_all = finetuned_encoder.encode(query_all)
    new_index_all = finetuned_encoder.encode(index_all)
    evaluator = Evaluator(query_data=new_query_all, index_data=new_index_all)
    ev = evaluator.evaluate()
    plot_metrics(ev, 'finetuned_m.png')

    # saving all the before-after images side-by-side
    save_before_after_image(path=f'preview-{data}.png')



import os
from typing import Dict, List, Tuple

from docarray import DocumentArray
from jina.helper import random_port
from now_common.preprocess import preprocess_images
from now_common.utils import common_setup, get_indexer_config

from now.apps.base.app import JinaNOWApp
from now.constants import CLIP_USES, EXTERNAL_CLIP_HOST, Apps, DemoDatasets, Modalities
from now.now_dataclasses import UserInput


class ImageToImage(JinaNOWApp):
    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> str:
        return Apps.IMAGE_TO_IMAGE

    @property
    def is_enabled(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return 'Image to image search app'

    @property
    def input_modality(self) -> Modalities:
        return Modalities.IMAGE

    @property
    def output_modality(self) -> Modalities:
        return Modalities.IMAGE

    @property
    def required_docker_memory_in_gb(self) -> int:
        return 8

    @property
    def finetune_datasets(self) -> [Tuple]:
        return (DemoDatasets.DEEP_FASHION, DemoDatasets.BIRD_SPECIES)

    def set_flow_yaml(self, **kwargs):
        finetuning = kwargs.get('finetuning', False)

        now_package_dir = os.path.abspath(
            os.path.join(__file__, '..', '..', '..', '..')
        )
        flow_dir = os.path.join(now_package_dir, 'now_common', 'flow')

        if finetuning:
            self.flow_yaml = os.path.join(flow_dir, 'ft-flow-clip.yml')
        else:
            self.flow_yaml = os.path.join(flow_dir, 'flow-clip.yml')

    @property
    def supported_file_types(self) -> List[str]:
        return ['gif', 'png', 'jpg', 'jpeg']

    def setup(
        self, dataset: DocumentArray, user_input: UserInput, kubectl_path
    ) -> Dict:
        indexer_config = get_indexer_config(len(dataset))
        is_remote = user_input.deployment_type == 'remote'
        encoder_with = {
            'ENCODER_HOST': EXTERNAL_CLIP_HOST if is_remote else '0.0.0.0',
            'ENCODER_PORT': 443 if is_remote else random_port(),
            'ENCODER_USES_TLS': True if is_remote else False,
            'ENCODER_IS_EXTERNAL': True if is_remote else False,
        }
        return common_setup(
            app_instance=self,
            user_input=user_input,
            dataset=dataset,
            encoder_uses=CLIP_USES[user_input.deployment_type][0],
            encoder_with=encoder_with,
            encoder_uses_with={
                'pretrained_model_name_or_path': CLIP_USES[user_input.deployment_type][
                    1
                ]
            },
            pre_trained_embedding_size=CLIP_USES[user_input.deployment_type][2],
            indexer_uses=indexer_config['indexer_uses'],
            kubectl_path=kubectl_path,
            indexer_resources=indexer_config['indexer_resources'],
        )

    def preprocess(
        self, da: DocumentArray, user_input: UserInput, is_indexing=False
    ) -> DocumentArray:
        return preprocess_images(da=da)



import io
import multiprocessing as mp

from docarray import Document, DocumentArray
from PIL import Image
from tqdm import tqdm


def to_thumbnail_jpg(doc: Document):
    if doc.tensor is not None:
        im = Image.fromarray(doc.tensor)
        im.thumbnail((256, 256))
        im = im.convert('RGB')
        doc.tensor = None
        img_byte_arr = io.BytesIO()
        im.save(img_byte_arr, format="JPEG", quality=75)
        doc.blob = img_byte_arr.getvalue()
    return doc


def convert_to_jpeg(dataset: str, num_workers: int = 8):
    path = f'{dataset}.bin'

    print(f'===> {dataset}')
    print(f'  Loading {dataset} dataset from {path} ...')
    docs = DocumentArray.load_binary(path, compress='gzip')
    print(f'  Dataset size: {len(docs)}')

    print(f' convert images to jpeg for smaller datasets')
    # build docs
    with mp.Pool(processes=num_workers) as pool:
        docs = list(tqdm(pool.imap(to_thumbnail_jpg, docs)))
    docs = DocumentArray(docs)

    print(f'  Saving converted docs ...')
    out = f'{dataset}.jpeg.bin'
    docs.save_binary(out)
    print(f'  Saved converted docs to {out} ...')


def main():
    """
    Main method.
    """
    datasets = [
        'deepfashion',
        'deepfashion.img10',
        'deepfashion.ViT-B32',
        'deepfashion.ViT-B16',
        'deepfashion.ViT-L14',
        'nih-chest-xrays',
        'nih-chest-xrays.img10',
        'nih-chest-xrays.ViT-B32',
        'nih-chest-xrays.ViT-B16',
        'nih-chest-xrays.ViT-L14',
        'geolocation-geoguessr',
        'geolocation-geoguessr.img10',
        'geolocation-geoguessr.ViT-B32',
        'geolocation-geoguessr.ViT-B16',
        'geolocation-geoguessr.ViT-L14',
        'stanford-cars',
        'stanford-cars.img10',
        'stanford-cars.ViT-B32',
        'stanford-cars.ViT-B16',
        'stanford-cars.ViT-L14',
        'bird-species',
        'bird-species.img10',
        'bird-species.ViT-B32',
        'bird-species.ViT-B16',
        'bird-species.ViT-L14',
        'best-artworks',
        'best-artworks.img10',
        'best-artworks.ViT-B32',
        'best-artworks.ViT-B16',
        'best-artworks.ViT-L14',
    ]

    for dataset in datasets:
        convert_to_jpeg(dataset)


if __name__ == '__main__':
    main()



import base64
import json
import os
import pkgutil
import shutil
import signal
import stat
import sys
import tempfile
from collections import namedtuple
from collections.abc import MutableMapping
from concurrent.futures import ThreadPoolExecutor
from os.path import expanduser as user
from typing import Dict, List, Optional, Union

import boto3
import hubble
import numpy as np
import yaml
from docarray import Document, DocumentArray
from PIL import Image, ImageDraw, ImageFont
from pyfiglet import Figlet
from rich.console import Console

from now.thirdparty.PyInquirer.prompt import prompt

colors = [
    "navy",
    "turquoise",
    "darkorange",
    "cornflowerblue",
    "teal",
    'maroon',
    'purple',
    'green',
    'lime',
    'navy',
    'teal',
]


def get_device():
    # return only cpu as we want all our processed to run on cpu
    return "cpu"  # "cuda" if torch.cuda.is_available() else "cpu"


def save_before_after_image(path):
    figs = [
        ['pretrained.png', 'finetuned.png'],
        # ['pretrained_pr.png', 'finetuned_pr.png'],
        ['pretrained_m.png', 'finetuned_m.png'],
    ]
    [im1, im2, im3, im4] = [Image.open(img) for imgs in figs for img in imgs]

    big_w = im1.width if im1.width > im3.width else im3.width

    im1 = im1.resize(
        (big_w, int(im1.height * big_w / im1.width)), resample=Image.BICUBIC
    )
    im2 = im2.resize(
        (big_w, int(im2.height * big_w / im2.width)), resample=Image.BICUBIC
    )
    im3 = im3.resize(
        (big_w, int(im3.height * big_w / im3.width)), resample=Image.BICUBIC
    )
    im4 = im4.resize(
        (big_w, int(im4.height * big_w / im4.width)), resample=Image.BICUBIC
    )
    # im5 = im5.resize(
    #     (big_w, int(im5.height * big_w / im5.width)), resample=Image.BICUBIC
    # )
    # im6 = im6.resize(
    #     (big_w, int(im6.height * big_w / im6.width)), resample=Image.BICUBIC
    # )

    width = max(im1.width + im2.width, im3.width + im4.width) + 200
    height = (
        max(
            im1.height + im3.height,
            im2.height + im4.height,
        )
        + 300
    )

    dst = Image.new('RGB', (width, height), color=(255, 255, 255))

    dst.paste(im1, (10, 150))
    dst.paste(im2, (im1.width + 150, 150))
    dst.paste(im3, (10, im1.height + 250))
    dst.paste(im4, (im1.width + 120, im1.height + 250))
    draw = ImageDraw.Draw(dst)
    font = ImageFont.truetype(
        'now/fonts/arial.ttf', 50
    )  # this font does not work in docker currently
    # font = ImageFont.load_default()
    draw.line(
        (
            im1.width + 80,
            0,
            im1.width + 80,
            im1.height + im2.height + im3.height + 250,
        ),
        width=10,
        fill='black',
    )  # vertical line
    draw.line(
        (0, im1.height + 200, im1.width + im2.width + 250, im1.height + 200),
        width=10,
        fill='black',
    )  # horizontal line
    draw.line(
        (
            0,
            im1.height + im2.height,
            im1.width + im2.width + 250,
            im1.height + im2.height,
        ),
        width=10,
        fill='black',
    )  # horizontal line
    draw.text(((im1.width - 150) // 2, 0), "Pre-trained Results", 0, font=font)
    draw.text(
        (im1.width + 50 + (im2.width // 2), 0),
        "Finetuned Results",
        0,
        font=font,
    )
    font = ImageFont.truetype('now/fonts/arial.ttf', 30)
    draw.text((20, 100), "Query", 0, font=font)
    draw.text((im1.width + 150, 100), "Query", 0, font=font)
    draw.text((im1.width // 2, 100), "Top-k", 0, font=font)
    draw.text((im1.width + 100 + (im2.width // 2), 100), "Top-k", 0, font=font)
    dst.show(title=path.split('.')[0])
    dst.save(path)
    for imgs in figs:
        for img in imgs:
            if os.path.isfile(img):
                os.remove(img)


def visual_result(
    data,
    querys,
    output: str = None,
    canvas_size: int = 1500,
    channel_axis: int = -1,
    img_per_row=11,
    label='finetuner_label',
    unique=False,
):
    img_size = int(canvas_size / img_per_row)
    img_h, img_w = img_size, img_size

    if data == 'geolocation-geoguessr':
        img_h, img_w = img_size, 2 * img_size

    sprite_img = np.zeros([img_h * len(querys), img_w * img_per_row, 3], dtype='uint8')

    for img_id, q in enumerate(querys):
        _d = Document(q, copy=True)
        if _d.content_type != 'tensor':
            _d.convert_blob_to_image_tensor()
            channel_axis = -1

        _d.set_image_tensor_channel_axis(channel_axis, -1).set_image_tensor_shape(
            shape=(img_h, img_w)
        )

        row_id = img_id
        col_id = 0
        sprite_img[
            (row_id * img_h) : ((row_id + 1) * img_h),
            (col_id * img_w) : ((col_id + 1) * img_w),
        ] = _d.tensor

        sprite_img[
            (row_id * img_h) : ((row_id + 1) * img_h),
            ((col_id + 1) * img_w) : ((col_id + 2) * img_w),
        ] = np.ones_like(_d.tensor.shape)

        for j, d in enumerate(q.matches, start=1):
            _d = Document(d, copy=True)
            if _d.content_type != 'tensor':
                _d.convert_blob_to_image_tensor()
                channel_axis = -1

            _d.set_image_tensor_channel_axis(channel_axis, -1).set_image_tensor_shape(
                shape=(img_h - 10, img_w - 10)
            )

            match_img = np.ones([img_h, img_w, 3], dtype='uint8')
            match_img[5:-5, 5:-5] = _d.tensor  # center the match results image
            if not unique:
                # apply green if it is same class else red
                if q.tags[label] == _d.tags[label]:
                    match_img[0:5, ...] = (0, 255, 0)
                    match_img[-5:-1, ...] = (0, 255, 0)
                    match_img[:, 0:5, ...] = (0, 255, 0)
                    match_img[:, -5:-1, ...] = (0, 255, 0)
                else:
                    match_img[0:5, ...] = (255, 0, 0)
                    match_img[-5:-1, ...] = (255, 0, 0)
                    match_img[:, 0:5, ...] = (255, 0, 0)
                    match_img[:, -5:-1, ...] = (255, 0, 0)
            else:
                match_img[0:5, ...] = (0, 0, 0)
                match_img[-5:-1, ...] = (0, 0, 0)
                match_img[:, 0:5, ...] = (0, 0, 0)
                match_img[:, -5:-1, ...] = (0, 0, 0)

            # paste it on the main canvas
            col_id = j + 1
            sprite_img[
                (row_id * img_h) : ((row_id + 1) * img_h),
                (col_id * img_w) : ((col_id + 1) * img_w),
            ] = match_img

    from PIL import Image

    img = Image.fromarray(sprite_img)
    if output:
        with open(output, 'wb') as fp:
            img.save(fp)


def plot_metrics(metrics_dict, title):
    dst = Image.new(
        'RGB',
        (500, 300),
        color=(255, 255, 255),
    )
    draw = ImageDraw.Draw(dst)
    font = ImageFont.truetype(
        'now/fonts/arial.ttf', 20
    )  # this font does not work in docker currently
    # font = ImageFont.load_default()
    for idx, (key, val) in enumerate(metrics_dict.items()):
        draw.text((50, idx * 30), f'{key}', 0, font=font)
        draw.text((250, idx * 30), f":  {val:.3f}", 0, font=font)

    dst.save(title)


def custom_spinner():
    SPINNERS_DATA = pkgutil.get_data(__name__, "custom-spinners.json").decode("utf-8")

    def _hook(dct):
        return namedtuple("Spinner", dct.keys())(*dct.values())

    return json.loads(SPINNERS_DATA, object_hook=_hook)


def download_file(path, r_raw):
    with path.open("wb") as f:
        shutil.copyfileobj(r_raw, f)


def download(url, filename):
    import functools
    import pathlib

    import requests
    from tqdm.auto import tqdm

    r = requests.get(url, stream=True, allow_redirects=True)
    if r.status_code != 200:
        r.raise_for_status()  # Will only raise for 4xx codes, so...
        raise RuntimeError(f"Request to {url} returned status code {r.status_code}")
    file_size = int(r.headers.get('Content-Length', 0))

    path = pathlib.Path(filename).expanduser().resolve()
    path.parent.mkdir(parents=True, exist_ok=True)

    desc = "(Unknown total file size)" if file_size == 0 else ""
    r.raw.read = functools.partial(
        r.raw.read, decode_content=True
    )  # Decompress if needed

    if 'NOW_CI_RUN' in os.environ:
        download_file(path, r.raw)
    else:
        with tqdm.wrapattr(r.raw, "read", total=file_size, desc=desc) as r_raw:
            download_file(path, r_raw)

    return path


def get_rich_console():
    """
    Function to get jina rich default console.
    :return: rich console
    """
    return Console(
        force_terminal=True, force_interactive=True
    )  # It forces render in any terminal, especily in PyCharm


def copytree(src, dst, symlinks=False, ignore=None):
    if not os.path.exists(dst):
        os.makedirs(dst)
        shutil.copystat(src, dst)
    lst = os.listdir(src)
    if ignore:
        excl = ignore(src, lst)
        lst = [x for x in lst if x not in excl]
    for item in lst:
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if symlinks and os.path.islink(s):
            if os.path.lexists(d):
                os.remove(d)
            os.symlink(os.readlink(s), d)
            try:
                st = os.lstat(s)
                mode = stat.S_IMODE(st.st_mode)
                os.lchmod(d, mode)
            except Exception:
                pass  # lchmod not available
        elif os.path.isdir(s):
            copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def my_handler(signum, frame, spinner):
    with spinner.hidden():
        sys.stdout.write("Program terminated!\n")
    spinner.stop()
    exit(0)


def flow_definition(dirpath) -> Dict:
    with open(dirpath) as f:
        return yaml.safe_load(f.read())


class BetterEnum:
    def __iter__(self):
        return [getattr(self, x) for x in dir(self) if ('__' not in x)].__iter__()


def to_camel_case(text):
    """Convert text to camel case in great coding style"""
    s = text.replace("_", " ")
    s = s.split()
    return ''.join(i.capitalize() for i in s)


sigmap = {signal.SIGINT: my_handler, signal.SIGTERM: my_handler}


def write_env_file(env_file, config):
    config_string = '\n'.join([f'{key}={value}' for key, value in config.items()])
    with open(env_file, 'w+') as fp:
        fp.write(config_string)


@hubble.login_required
def jina_auth_login():
    pass


def get_info_hubble(user_input):
    with open(user('~/.jina/config.json')) as fp:
        config_val = json.load(fp)
        user_token = config_val['auth_token']
    client = hubble.Client(token=user_token, max_retries=None, jsonify=True)
    response = client.get_user_info()
    user_input.admin_emails = (
        [response['data']['email']] if 'email' in response['data'] else []
    )
    if not user_input.admin_emails:
        print(
            'Your hubble account is not verified. Please verify your account to deploy your flow as admin.'
        )
    user_input.jwt = {'token': user_token}
    return response['data'], user_token


def print_headline():
    f = Figlet(font='slant')
    print('Welcome to:')
    print(f.renderText('Jina NOW'))
    print('Get your search case up and running - end to end.\n')
    print(
        'You can choose between image and text search. \nJina NOW trains a model, pushes it to Jina Hub '
        'and deploys a Flow and a playground app in the cloud or locally. \nCheck out one of our demos or bring '
        'your own data.\n'
    )
    print('If you want learn more about our framework please visit docs.jina.ai')
    print(
        '💡 Make sure you give enough memory to your Docker daemon. '
        '5GB - 8GB should be okay.'
    )
    print()


def maybe_prompt_user(questions, attribute, **kwargs):
    """
    Checks the `kwargs` for the `attribute` name. If present, the value is returned directly.
    If not, the user is prompted via the cmd-line using the `questions` argument.

    :param questions: A dictionary that is passed to `PyInquirer.prompt`
        See docs: https://github.com/CITGuru/PyInquirer#documentation
    :param attribute: Name of the value to get. Make sure this matches the name in `kwargs`

    :return: A single value of either from `kwargs` or the user cli input.
    """
    if kwargs and kwargs.get(attribute) is not None:
        return kwargs[attribute]
    else:
        answer = prompt(questions)
        return answer[attribute]


def _prompt_value(
    name: str,
    prompt_message: str,
    prompt_type: str = 'input',
    choices: Optional[List[Union[Dict, str]]] = None,
    **kwargs: Dict,
):
    qs = {'name': name, 'type': prompt_type, 'message': prompt_message}

    if choices is not None:
        qs['choices'] = choices
        qs['type'] = 'list'
    return maybe_prompt_user(qs, name, **kwargs)


def _maybe_download_from_s3(
    docs: DocumentArray, tmpdir: tempfile.TemporaryDirectory, user_input, max_workers
):
    """Downloads file to local temporary dictionary, saves S3 URI to `tags['uri']` and modifies `uri` attribute of
    document to local path in-place.

    :param doc: document containing URI pointing to the location on S3 bucket
    :param tmpdir: temporary directory in which files will be saved
    :param user_input: User iput which contain aws credentials
    :param max_workers: number of threads to create in the threadpool executor to make execution faster
    """

    def download(bucket, uri):
        path_s3 = '/'.join(uri.split('/')[3:])
        path_local = os.path.join(
            str(tmpdir),
            base64.b64encode(bytes(path_s3, "utf-8")).decode("utf-8"),
        )
        bucket.download_file(
            path_s3,
            path_local,
        )
        return path_local

    def convert_fn(d: Document) -> Document:
        """Downloads files and tags from S3 bucket and updates the content uri and the tags uri to the local path"""
        d.tags['uri'] = d.uri
        session = boto3.session.Session(
            aws_access_key_id=user_input.aws_access_key_id,
            aws_secret_access_key=user_input.aws_secret_access_key,
            region_name=user_input.aws_region_name,
        )
        bucket = session.resource('s3').Bucket(d.uri.split('/')[2])
        d.uri = download(bucket=bucket, uri=d.uri)
        if 'tag_uri' in d.tags:
            d.tags['tag_uri'] = download(bucket, d.tags['tag_uri'])
            with open(d.tags['tag_uri'], 'r') as fp:
                tags = json.load(fp)
                tags = flatten_dict(tags)
                d.tags.update(tags)
            del d.tags['tag_uri']
        return d

    docs_to_download = []
    for d in docs:
        if d.uri.startswith('s3://'):
            docs_to_download.append(d)

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for d in docs_to_download:
            f = executor.submit(convert_fn, d)
            futures.append(f)
        for f in futures:
            f.result()


def flatten_dict(d, parent_key='', sep='_'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.
"""
from __future__ import annotations, print_function, unicode_literals

import inspect
import pathlib

from now_common import options

import now.utils
from now.apps.base.app import JinaNOWApp
from now.now_dataclasses import DialogOptions, UserInput

cur_dir = pathlib.Path(__file__).parent.resolve()


def configure_app(**kwargs) -> JinaNOWApp:
    now.utils.print_headline()
    app_name = configure_option(options.APP, None, **kwargs)
    return options._construct_app(app_name)


def configure_user_input(app_instance: JinaNOWApp, **kwargs) -> UserInput:
    user_input = UserInput()
    user_input.app_instance = app_instance
    # Ask the base/common options
    for option in options.base_options:
        configure_option(option, user_input, **kwargs)
    # Ask app specific options
    for option in user_input.app_instance.options:
        configure_option(option, user_input, **kwargs)

    return user_input


def configure_option(option: DialogOptions, user_input: UserInput, **kwargs):
    # If there is any pre function then invoke that - commonly used to fill choices if needed
    # if inspect.isfunction(option.pre_func):
    #     option.post_func(user_input, option, **kwargs)

    # Check if it is dependent on some other dialog options
    if option.depends_on and not option.conditional_check(user_input):
        return

    # Populate choices if needed
    if option.choices and inspect.isfunction(option.choices):
        option.choices = option.choices(user_input, **kwargs)

    val = now.utils._prompt_value(
        **option.__dict__,
        **kwargs,
    )

    if hasattr(user_input, option.name):
        setattr(user_input, option.name, val)

    # If there is any post function then invoke that
    kwargs[option.name] = val
    if inspect.isfunction(option.post_func):
        option.post_func(user_input, **kwargs)

    return val



import os
from typing import Dict, List

import cowsay
from docarray import Document, DocumentArray
from now_common.utils import common_setup

from now.apps.base.app import JinaNOWApp
from now.constants import NOW_ANNLITE_INDEXER_VERSION, Apps, DemoDatasets, Modalities
from now.deployment.deployment import which
from now.now_dataclasses import UserInput


class MusicToMusic(JinaNOWApp):
    """
    Music2Music search.

    Pre-trained head weights are stored here
    https://console.cloud.google.com/storage/browser/jina-fashion-data/model/music?project=jina-simpsons-florian
    To re-built, go in now/hub/head_encoder with place the model weights in this folder
    and run "jina hub push --private . -t linear_head_encoder_music_2k"
    """

    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> str:
        return Apps.MUSIC_TO_MUSIC

    @property
    def is_enabled(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return 'Music to music search app'

    @property
    def input_modality(self) -> Modalities:
        return Modalities.MUSIC

    @property
    def output_modality(self) -> Modalities:
        return Modalities.MUSIC

    @property
    def required_docker_memory_in_gb(self) -> int:
        return 10

    def set_flow_yaml(self, **kwargs):
        finetuning = kwargs.get('finetuning', False)
        demo_data = kwargs.get('demo_data', False)
        if finetuning + demo_data > 1:
            raise ValueError(
                f"Can't set flow to more than one mode but have "
                f"demo_data={demo_data}, finetuning={finetuning}"
            )

        flow_dir = os.path.abspath(os.path.join(__file__, '..'))

        if demo_data:
            self.flow_yaml = os.path.join(flow_dir, 'demo-data-flow-music.yml')
        else:
            self.flow_yaml = os.path.join(flow_dir, 'ft-flow-music.yml')

    @property
    def supported_file_types(self) -> List[str]:
        return ['mp3']

    def _check_requirements(self) -> bool:
        if not ffmpeg_is_installed():
            _handle_ffmpeg_install_required()
            return False
        return True

    def setup(
        self, dataset: DocumentArray, user_input: UserInput, kubectl_path
    ) -> Dict:
        # needed to write a custom solution for music 2 music app as we need to allow to integrate
        # externally pretrained executors for the demo datasets
        pre_trained_head_map = {
            DemoDatasets.MUSIC_GENRES_ROCK: 'FinetunedLinearHeadEncoderMusicRock',
            DemoDatasets.MUSIC_GENRES_MIX: 'FineTunedLinearHeadEncoderMusicMix',
        }

        # will execute finetuning on custom datasets (if possible) but not for demo datasets
        env_dict = common_setup(
            app_instance=self,
            user_input=user_input,
            dataset=dataset,
            encoder_uses='BiModalMusicTextEncoderV2',
            encoder_uses_with={},
            indexer_uses=f'NOWAnnLiteIndexer/v{NOW_ANNLITE_INDEXER_VERSION}',
            kubectl_path=kubectl_path,
            indexer_resources={},
            pre_trained_embedding_size=512,
        )

        # can reuse large part of other code but need to make some adjustments
        if user_input.data in pre_trained_head_map:
            print(f'⚡️ Using cached hub model for speed')

            env_dict[
                'LINEAR_HEAD_NAME'
            ] = f"jinahub+docker://{pre_trained_head_map[user_input.data]}"

            self.set_flow_yaml(demo_data=True)

        return env_dict

    def preprocess(
        self, da: DocumentArray, user_input: UserInput, is_indexing=False
    ) -> DocumentArray:
        from pydub import AudioSegment

        def convert_fn(d: Document):
            try:
                if d.blob == b'':
                    if d.uri:
                        if d.uri.startswith(f'data:{d.mime_type}'):
                            d.load_uri_to_blob()
                        else:
                            AudioSegment.from_file(d.uri)  # checks if file is valid
                            with open(d.uri, 'rb') as fh:
                                d.blob = fh.read()
                return d
            except Exception as e:
                return d

        for d in da:
            convert_fn(d)
        return DocumentArray(d for d in da if d.blob != b'')


def ffmpeg_is_installed():
    return which("ffmpeg")


def _handle_ffmpeg_install_required():
    bc_red = '\033[91m'
    bc_end = '\033[0m'
    print()
    print(
        f"{bc_red}To use the audio modality you need the ffmpeg audio processing"
        f" library installed on your system.{bc_end}"
    )
    print(
        f"{bc_red}For MacOS please run 'brew install ffmpeg' and on"
        f" Linux 'apt-get install ffmpeg libavcodec-extra'.{bc_end}"
    )
    print(
        f"{bc_red}After the installation, restart Jina Now and have fun with music search 🎸!{bc_end}"
    )
    cowsay.cow('see you soon 👋')
    exit(1)



import json
import os
import tempfile
from copy import deepcopy
from os.path import expanduser as user
from typing import Dict, List, Optional

import hubble
from docarray import Document, DocumentArray
from jina import __version__ as jina_version

from now.apps.base.app import JinaNOWApp
from now.constants import (
    NOW_ANNLITE_INDEXER_VERSION,
    NOW_PREPROCESSOR_VERSION,
    PREFETCH_NR,
    DatasetTypes,
)
from now.finetuning.run_finetuning import finetune
from now.finetuning.settings import FinetuneSettings, parse_finetune_settings
from now.now_dataclasses import UserInput
from now.utils import _maybe_download_from_s3


def common_get_flow_env_dict(
    finetune_settings: FinetuneSettings,
    encoder_uses: str,
    encoder_with: Dict,
    encoder_uses_with: Dict,
    pre_trained_embedding_size: int,
    indexer_uses: str,
    indexer_resources: Dict,
    user_input: UserInput,
    tags: List,
):
    """Returns dictionary for the environments variables for the clip & music flow.yml files."""
    if (
        finetune_settings.perform_finetuning and finetune_settings.bi_modal
    ) or user_input.app_instance.app_name == 'music_to_music':
        pre_trained_embedding_size = pre_trained_embedding_size * 2

    config = {
        'JINA_VERSION': jina_version,
        'ENCODER_NAME': f'jinahub+docker://{encoder_uses}',
        'N_DIM': finetune_settings.finetune_layer_size
        if finetune_settings.perform_finetuning
        or user_input.app_instance.app_name == 'music_to_music'
        else pre_trained_embedding_size,
        'PRE_TRAINED_EMBEDDINGS_SIZE': pre_trained_embedding_size,
        'INDEXER_NAME': f'jinahub+docker://{indexer_uses}',
        'PREFETCH': PREFETCH_NR,
        'PREPROCESSOR_NAME': f'jinahub+docker://NOWPreprocessor/v{NOW_PREPROCESSOR_VERSION}',
        'APP': user_input.app_instance.app_name,
        'COLUMNS': tags,
        'ADMIN_EMAILS': user_input.admin_emails or [] if user_input.secured else [],
        'USER_EMAILS': user_input.user_emails or [] if user_input.secured else [],
        **encoder_with,
        **indexer_resources,
    }
    if encoder_uses_with.get('pretrained_model_name_or_path'):
        config['PRE_TRAINED_MODEL_NAME'] = encoder_uses_with[
            "pretrained_model_name_or_path"
        ]
    if finetune_settings.perform_finetuning:
        config['FINETUNE_ARTIFACT'] = finetune_settings.finetuned_model_artifact
        config['JINA_TOKEN'] = finetune_settings.token

    # retention days
    if 'NOW_CI_RUN' in os.environ:
        config[
            'RETENTION_DAYS'
        ] = 0  # JCloud will delete after 24hrs of being idle if not deleted in CI
    else:
        config['RETENTION_DAYS'] = 7  # for user deployment set it to 30 days

    return config


def common_setup(
    app_instance: JinaNOWApp,
    user_input: UserInput,
    dataset: DocumentArray,
    encoder_uses: str,
    encoder_uses_with: Dict,
    indexer_uses: str,
    pre_trained_embedding_size: int,
    kubectl_path: str,
    encoder_with: Optional[Dict] = {},
    indexer_resources: Optional[Dict] = {},
) -> Dict:
    # should receive pre embedding size
    finetune_settings = parse_finetune_settings(
        pre_trained_embedding_size=pre_trained_embedding_size,
        user_input=user_input,
        dataset=dataset,
        finetune_datasets=app_instance.finetune_datasets,
        model_name='mlp',
        add_embeddings=True,
        loss='TripletMarginLoss',
    )
    tags = _extract_tags_annlite(deepcopy(dataset[0]), user_input)
    env_dict = common_get_flow_env_dict(
        finetune_settings=finetune_settings,
        encoder_uses=encoder_uses,
        encoder_with=encoder_with,
        encoder_uses_with=encoder_uses_with,
        pre_trained_embedding_size=pre_trained_embedding_size,
        indexer_uses=indexer_uses,
        indexer_resources=indexer_resources,
        user_input=user_input,
        tags=tags,
    )

    if finetune_settings.perform_finetuning:
        try:
            artifact_id, token = finetune(
                finetune_settings=finetune_settings,
                app_instance=app_instance,
                dataset=dataset,
                user_input=user_input,
                env_dict=env_dict,
                kubectl_path=kubectl_path,
            )

            finetune_settings.finetuned_model_artifact = artifact_id
            finetune_settings.token = token

            env_dict['FINETUNE_ARTIFACT'] = finetune_settings.finetuned_model_artifact
            env_dict['JINA_TOKEN'] = finetune_settings.token
        except Exception as e:
            print(
                'Finetuning is currently offline. The programm execution still continues without finetuning. Please report the following exception to us:'
            )
            import traceback

            traceback.print_exc()
            finetune_settings.perform_finetuning = False

    app_instance.set_flow_yaml(
        finetuning=finetune_settings.perform_finetuning, dataset_len=len(dataset)
    )

    return env_dict


def _get_email():
    try:
        with open(user('~/.jina/config.json')) as fp:
            config_val = json.load(fp)
            user_token = config_val['auth_token']
            client = hubble.Client(token=user_token, max_retries=None, jsonify=True)
            response = client.get_user_info()
        if 'email' in response['data']:
            return response['data']['email']
        return ''
    except FileNotFoundError:
        return ''


def get_indexer_config(num_indexed_samples: int) -> Dict:
    """Depending on the number of samples, which will be indexed, indexer and its resources are determined.

    :param num_indexed_samples: number of samples which will be indexed; should incl. chunks for e.g. text-to-video app
    """
    config = {'indexer_uses': f'NOWAnnLiteIndexer/v{NOW_ANNLITE_INDEXER_VERSION}'}
    threshold1 = 250_000
    if num_indexed_samples <= threshold1:
        config['indexer_resources'] = {'INDEXER_CPU': 0.1, 'INDEXER_MEM': '2G'}
    else:
        config['indexer_resources'] = {'INDEXER_CPU': 1.0, 'INDEXER_MEM': '4G'}

    return config


def _extract_tags_annlite(d: Document, user_input):
    print(
        'We assume all tags follow the same structure, only first json file will be used to determine structure'
    )
    with tempfile.TemporaryDirectory() as tmpdir:
        if user_input and user_input.custom_dataset_type == DatasetTypes.S3_BUCKET:
            _maybe_download_from_s3(
                docs=DocumentArray([d]),
                tmpdir=tmpdir,
                user_input=user_input,
                max_workers=1,
            )
    tags = set()
    for tag, _ in d.tags.items():
        tags.add((tag, str(tag.__class__.__name__)))
    final_tags = [list(tag) for tag in tags]
    return final_tags



import os
from typing import Dict, List, Optional, Union, Set

from docarray import Document, DocumentArray
from PIL import Image

from .connector import ElasticsearchConnector

import logging

logging.getLogger("PIL.Image").setLevel(logging.CRITICAL + 1)


ID_TAG = 'id'
FIELD_TAG = 'field_name'
EXTRACTION_TYPE_TAG = 'extraction_type'


class ElasticsearchExtractor:
    def __init__(
        self,
        query: Dict,
        index: str,
        connection_str: str,
        connection_args: Optional[Dict] = None,
    ):
        """
        For extracting documents from Elasticsearch into a `docarray.DocumentArray`
        dataset, this class implements an iterator which yields `docarray.Document`
        objects. To specify the data for extraction, one needs to provide an
        es query together with the index name and parameters to connect to
        the Elasticsearch instance.
        :param query: Elasticsearch query in the form of a JSON string
        :param index: Name of the ES index containing the documents to be extracted
        :param connection_str: A connection string for the ES instance. Usually, it
            includes url, port, username, password, etc. Typically, it has the form:
            'https://{user_name}:{password}@{host}:{port}'
        :param connection_args: Dictionary with additional connection arguments,
            e.g., information about certificates
        """
        self._es_connector = ElasticsearchConnector(
            connection_str=connection_str,
            connection_args=(connection_args if connection_args else {}),
        )
        self._query = query
        self._index = index
        self._document_cache = []
        self._query_result = self._es_connector.get_documents_by_query(
            self._query, self._index
        )
        self._supported_pil_extensions = self._get_supported_image_extensions()

    def extract(self) -> DocumentArray:
        """
        Returns extracted data as a `DocumentArray` where every `Document`
        contains chunks for each field.
        For Example:
        Document(chunks=[
            Document(content='hello', modality='text', tags={'field_name': 'title'}),
            Document(content='https://bla.com/img.jpeg', modality='image', tags={'field_name': 'uris'}),
            ]
        )
        """
        return DocumentArray(
            [self._transform_es_doc(doc) for doc in self._extract_documents()]
        )

    def _extract_documents(self):
        try:
            next_doc = self._get_next_document()
            while next_doc:
                yield next_doc
                next_doc = self._get_next_document()
        except StopIteration:
            self._es_connector.close()
            return

    @classmethod
    def _transform_es_doc(cls, document: Document) -> Document:
        """
        Transform data extracted from Elasticsearch to a more convenient form.
        :param document: `Document` containing ES data.
        :return: Transformed `Document`.
        """
        attr_values, attr_modalities = {}, {}
        cls._parse_es_doc(document, attr_values, attr_modalities, [])
        transformed_doc = Document(
            chunks=[
                Document(
                    content=attr_values[name],
                    modality=attr_modalities[name],
                    tags={'field_name': name},
                )
                for name in attr_values
            ]
        )
        return transformed_doc

    @classmethod
    def _parse_es_doc(
        cls,
        document: Document,
        attr_values: Dict,
        attr_modalities: Dict,
        names: List[str],
    ):
        """
        Extract attributes and modalities from a `Document` and store it as a dictionary.
        Recursively iterates across different chunks of the `Document` and collects
        attributes with their corresponding values.
        :param document: `Document` we want to transform.
        :param attr_values: Dictionary of attribute values extracted from the document.
        :param attr_modalities: Dictionary of attribute modalities extracted from the document.
        :param names: Name of an attribute (attribute names may be nested, e.g.
            info.cars, and we need to store name(s) on every level of recursion).
        """
        if not document.chunks:
            names.append(document.tags['field_name'])
            attr_name = '.'.join(names)
            attr_val = document.text if document.modality == 'text' else document.uri
            if attr_name not in attr_modalities:
                attr_modalities[attr_name] = document.modality
                attr_values[attr_name] = []
            attr_values[attr_name].append(attr_val)
        else:
            if 'field_name' in document.tags:
                names.append(document.tags['field_name'])
            for doc in document.chunks:
                cls._parse_es_doc(doc, attr_values, attr_modalities, names[:])

    def _get_next_document(self) -> Union[Document, None]:
        """
        Returns the next document from the Elasticsearch database.
        In order to retrieve further documents, Elasticsearch documents are retrieved
        in pages of multiple documents. After retrieving a page, its contained
        documents are stored in a document cache. If documents are left in the cache
        this function returns of of those documents. Otherwise, the next page is
        queried. If there is no page left, None will be returned.
        :return: extracted document
        """
        if len(self._document_cache) == 0:
            self._document_cache = next(self._query_result)
            if len(self._document_cache) == 0:
                return None
        return self._construct_document(self._document_cache.pop())

    def _construct_simple_document(self, key, content) -> Document:
        """
        Transforms the values of a filed of an Elasticsearch document into a
        `docarray.Document` object (without chunks).
        The name of the field where the attribute is extracted from is added as a tag
        to the document.
        :param key: the name of the field which stores the attribute in the
            Elasticsearch database
        :param content: the content of the attribute
        :return: `docarray.DocumentArray` object.
        """
        if os.path.splitext(content)[-1] in self._supported_pil_extensions:
            return Document(
                uri=content,
                modality='image',
                tags={
                    FIELD_TAG: key,
                    EXTRACTION_TYPE_TAG: 'literal',
                },
            )
        else:
            return Document(
                text=content,
                modality='text',
                tags={
                    FIELD_TAG: key,
                    EXTRACTION_TYPE_TAG: 'literal',
                },
            )

    def _construct_document(self, es_document: Dict) -> Document:
        """
        Constructs a `docarray.Document` from an Elasticsearch document returned by the
        `ElasticsearchConnector.get_documents_by_query` function.
        :param es_document: dictionary containing the content of an Elasticsearch
            document
        :return: A `docarray.Document` object where each chunk represents an element of
            a field from an Elasticsearch document.
        """
        root = (es_document, Document())
        frontier = [root]
        while len(frontier) > 0:
            new_frontier = []
            for es_doc, da_doc in frontier:
                for key in es_doc.keys():
                    new_doc = None
                    if key == 'id':
                        # process the id key value pair
                        da_doc.tags[ID_TAG] = es_doc[key]
                        da_doc.tags[EXTRACTION_TYPE_TAG] = 'root'
                    elif type(es_doc[key][0]) == dict:
                        # process a nested field
                        new_doc = Document(
                            tags={FIELD_TAG: key, EXTRACTION_TYPE_TAG: 'record'}
                        )
                        new_frontier.extend(
                            [(record, new_doc) for record in es_doc[key]]
                        )
                    elif type(es_doc[key][0]) == str:
                        # process a field with a list of string values
                        deep_chunks = [
                            self._construct_simple_document(key, content)
                            for content in es_doc[key]
                        ]
                        new_doc = Document(
                            chunks=deep_chunks, tags={EXTRACTION_TYPE_TAG: 'list'}
                        )
                    else:
                        raise RuntimeError('Unsupported type: ', type(es_doc[key][0]))
                    if new_doc is not None:
                        da_doc.chunks.append(new_doc)
            frontier = new_frontier
        return root[1]

    @staticmethod
    def _get_supported_image_extensions() -> Set[str]:
        extensions = Image.registered_extensions()
        return {ex for ex, f in extensions.items() if f in Image.OPEN}



# -*- coding: utf-8 -*-
"""
`expand` type question
"""
from prompt_toolkit.application import Application
from prompt_toolkit.filters import IsDone
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import LayoutDimension as D

from ..separator import Separator
from . import PromptParameterException
from .common import default_style, if_mousedown

# custom control based on FormattedTextControl


class InquirerControl(FormattedTextControl):
    def __init__(self, choices, default=None, **kwargs):
        self.pointer_index = 0
        self.answered = False
        self._init_choices(choices, default)
        self._help_active = False  # help is activated via 'h' key
        super().__init__(self._get_choice_tokens, **kwargs)

    def _init_choices(self, choices, default=None):
        # helper to convert from question format to internal format

        self.choices = []  # list (key, name, value)

        for i, c in enumerate(choices):
            if isinstance(c, Separator):
                self.choices.append(c)
            else:
                if isinstance(c, str):
                    self.choices.append((key, c, c))
                else:
                    key = c.get('key')
                    name = c.get('name')
                    value = c.get('value', name)
                    self.choices.append([key, name, value])

        # append the help choice
        self.choices.append(['h', 'Help, list all options', '__HELP__'])

        # set the default
        for i, choice in enumerate(self.choices):
            if isinstance(choice, list):
                key = choice[0]
                default = default or "h"
                if default == key:
                    self.pointer_index = i
                    choice[0] = key.upper()  # default key is in uppercase

    @property
    def choice_count(self):
        return len(self.choices)

    def _get_choice_tokens(self):
        tokens = []

        def _append(index, line):
            if isinstance(line, Separator):
                tokens.append(('class:separator', '   %s\n' % line))
            else:
                key = line[0]
                line = line[1]
                pointed_at = index == self.pointer_index

                @if_mousedown
                def select_item(mouse_event):
                    # bind option with this index to mouse event
                    self.pointer_index = index

                if pointed_at:
                    tokens.append(
                        ('class:selected', '  %s) %s' % (key, line), select_item)
                    )
                else:
                    tokens.append(('', '  %s) %s' % (key, line), select_item))
                tokens.append(('', '\n'))

        if self._help_active:
            # prepare the select choices
            for i, choice in enumerate(self.choices):
                _append(i, choice)
            tokens.append(('', '  Answer: %s' % self.choices[self.pointer_index][0]))
        else:
            tokens.append(('class:pointer', '>> '))
            tokens.append(('', self.choices[self.pointer_index][1]))
        return tokens

    def get_selected_value(self):
        # get value not label
        return self.choices[self.pointer_index][2]


def question(message, **kwargs):
    # TODO extract common parts for list, checkbox, rawlist, expand
    # TODO up, down navigation
    if not 'choices' in kwargs:
        raise PromptParameterException('choices')

    choices = kwargs.pop('choices', None)
    default = kwargs.pop('default', None)
    qmark = kwargs.pop('qmark', '?')
    # TODO style defaults on detail level
    style = kwargs.pop('style', default_style)

    ic = InquirerControl(choices, default)

    def get_prompt_tokens():
        tokens = []

        tokens.append(('class:questionmark', qmark))
        tokens.append(('class:question', ' %s ' % message))
        if not ic.answered:
            tokens.append(
                (
                    'class:instruction',
                    ' (%s)'
                    % ''.join(
                        [k[0] for k in ic.choices if not isinstance(k, Separator)]
                    ),
                )
            )
        else:
            tokens.append(('class:answer', ' %s' % ic.get_selected_value()))
        return tokens

    # @Condition
    # def is_help_active():
    #    return ic._help_active

    # assemble layout
    layout = HSplit(
        [
            Window(height=D.exact(1), content=FormattedTextControl(get_prompt_tokens)),
            ConditionalContainer(
                Window(ic),
                # filter=is_help_active & ~IsDone()  # ~ bitwise inverse
                filter=~IsDone(),  # ~ bitwise inverse
            ),
        ]
    )

    # key bindings
    kb = KeyBindings()

    @kb.add('c-q', eager=True)
    @kb.add('c-c', eager=True)
    def _(event):
        raise KeyboardInterrupt()

    # add key bindings for choices
    for i, c in enumerate(ic.choices):
        if not isinstance(c, Separator):

            def _reg_binding(i, keys):
                # trick out late evaluation with a "function factory":
                # http://stackoverflow.com/questions/3431676/creating-functions-in-a-loop
                @kb.add(keys, eager=True)
                def select_choice(event):
                    ic.pointer_index = i

            if c[0] not in ['h', 'H']:
                _reg_binding(i, c[0])
                if c[0].isupper():
                    _reg_binding(i, c[0].lower())

    @kb.add('H', eager=True)
    @kb.add('h', eager=True)
    def help_choice(event):
        ic._help_active = not ic._help_active

    @kb.add('enter', eager=True)
    def set_answer(event):
        selected_value = ic.get_selected_value()
        if selected_value == '__HELP__':
            ic._help_active = True
        else:
            ic.answered = True
            event.app.exit(result=selected_value)

    return Application(
        layout=Layout(layout), key_bindings=kb, mouse_support=True, style=style
    )



from typing import List

from docarray import Document, DocumentArray
from fastapi import APIRouter

from deployment.bff.app.v1.models.image import NowImageSearchRequestModel
from deployment.bff.app.v1.models.text import (
    NowTextIndexRequestModel,
    NowTextResponseModel,
)
from deployment.bff.app.v1.routers.helper import jina_client_post, process_query

router = APIRouter()


# Index
@router.post(
    "/index",
    summary='Add more text data to the indexer',
)
def index(data: NowTextIndexRequestModel):
    """
    Append the list of text to the indexer.
    """
    index_docs = DocumentArray()
    for text, uri, tags in zip(data.texts, data.uris, data.tags):
        if bool(text) + bool(uri) != 1:
            raise ValueError(f'Can only set one value but have text={text}, uri={uri}')
        if text:
            index_docs.append(Document(text=text, tags=tags))
        else:
            index_docs.append(Document(uri=uri, tags=tags))

    jina_client_post(
        data=data,
        inputs=index_docs,
        parameters={},
        endpoint='/index',
    )


# Search
@router.post(
    "/search",
    response_model=List[NowTextResponseModel],
    summary='Search text data via image as query',
)
def search(data: NowImageSearchRequestModel):
    """
    Retrieve matching text for a given image query. Image query should be
    `base64` encoded using human-readable characters - `utf-8`.
    """
    query_doc, filter_query = process_query(
        blob=data.image, uri=data.uri, conditions=data.filters
    )

    docs = jina_client_post(
        data=data,
        inputs=query_doc,
        parameters={'limit': data.limit, 'filter': filter_query},
        endpoint='/search',
    )

    return docs[0].matches.to_dict()



from abc import ABC, abstractmethod
from itertools import chain, combinations, permutations
from typing import Any, Dict, List

from docarray import Document, DocumentArray


class GeneratorFunction(ABC):
    @abstractmethod
    def process(self, document: Document) -> DocumentArray:
        """Generation function that produces a `DocumentArray` given a `Document`."""
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        """Name for the task config."""
        pass


class ImageNormalizer(GeneratorFunction):
    def __init__(self, scope: List[str]):
        """
        Class for image processing and normalization.

        :param scope: List of fields to consider for processing.
        """
        self._scope = scope

    def process(self, document: Document) -> DocumentArray:
        """
        Processes a `Document` containing image information.

        :param document: Document containing image data (uris).
        :return: `DocumentArray` of processed images.
        """
        normalized_imgs = DocumentArray()
        for chunk in document.chunks:
            if chunk.tags['field_name'] in self._scope:
                normalized_imgs.extend([Document(uri=uri) for uri in chunk.content])
        normalized_imgs.apply(self._process)
        return normalized_imgs

    @staticmethod
    def _process(doc: Document) -> Document:
        """Loads uri as image and applies normalization."""
        doc.load_uri_to_image_tensor().set_image_tensor_normalization()
        return doc

    @staticmethod
    def name() -> str:
        return 'image_normalization'


class TextProcessor(GeneratorFunction):
    def __init__(self, scope: List[str], powerset: bool = False, permute: bool = False):
        """
        Class for text processing with subsets and permutations.

        :param scope: List of fields to consider for processing.
        :param powerset: Creates subsets of text fields if set to `True`.
        :param permute: Creates permutations of fields if set to `True`.
        """
        self._scope = scope
        self._powerset = powerset
        self._permute = permute

    def process(self, document: Document) -> DocumentArray:
        """
        Processes a `Document` containing text information.

        :param document: Document containing text data.
        :param scope: List of fields to consider for processing.
        :return: `DocumentArray` of processed (and concatenated) text data.
        """
        document = {
            chunk.tags['field_name']: chunk.content for chunk in document.chunks
        }
        if self._powerset:
            key_sets = chain.from_iterable(
                combinations(self._scope, r) for r in range(1, len(self._scope) + 1)
            )
        else:
            key_sets = [self._scope]

        attribute_sets = [[document[key] for key in key_set] for key_set in key_sets]
        attribute_sets = [
            [item for sublist in attr_set for item in sublist]
            for attr_set in attribute_sets
        ]

        if self._permute:
            return DocumentArray(
                [
                    Document(content=' '.join(attr_seq), modality='text')
                    for attributes in attribute_sets
                    for attr_seq in permutations(attributes)
                ]
            )
        else:
            return DocumentArray(
                [
                    Document(content=' '.join(attributes), modality='text')
                    for attributes in attribute_sets
                ]
            )

    @staticmethod
    def name() -> str:
        return 'text_processing'



import base64
import os
import pathlib
import pickle
from os.path import join as osp

from docarray import DocumentArray

from now.constants import BASE_STORAGE_URL, DEMO_DATASET_DOCARRAY_VERSION, Modalities
from now.utils import download


def _fetch_da_from_url(
    url: str, downloaded_path: str = '~/.cache/jina-now'
) -> DocumentArray:
    data_dir = os.path.expanduser(downloaded_path)
    if not os.path.exists(osp(data_dir, 'data/tmp')):
        os.makedirs(osp(data_dir, 'data/tmp'))
    data_path = (
        data_dir
        + f"/data/tmp/{base64.b64encode(bytes(url, 'utf-8')).decode('utf-8')}.bin"
    )
    if not os.path.exists(data_path):
        download(url, data_path)

    try:
        da = DocumentArray.load_binary(data_path)
    except pickle.UnpicklingError:
        path = pathlib.Path(data_path).expanduser().resolve()
        os.remove(path)
        download(url, data_path)
        da = DocumentArray.load_binary(data_path)
    return da


def get_dataset_url(dataset: str, output_modality: Modalities) -> str:
    data_folder = None
    docarray_version = DEMO_DATASET_DOCARRAY_VERSION
    if output_modality == Modalities.IMAGE:
        data_folder = 'jpeg'
    elif output_modality == Modalities.TEXT:
        data_folder = 'text'
    elif output_modality == Modalities.MUSIC:
        data_folder = 'music'
    elif output_modality == Modalities.VIDEO:
        data_folder = 'video'
    elif output_modality == Modalities.TEXT_AND_IMAGE:
        data_folder = 'text-image'
    if output_modality not in [
        Modalities.MUSIC,
        Modalities.VIDEO,
        Modalities.TEXT_AND_IMAGE,
    ]:
        model_name = 'ViT-B32'
        return f'{BASE_STORAGE_URL}/{data_folder}/{dataset}.{model_name}-{docarray_version}.bin'
    else:
        return f'{BASE_STORAGE_URL}/{data_folder}/{dataset}-{docarray_version}.bin'



import os
from argparse import Namespace
from concurrent.futures import ProcessPoolExecutor

from now.cli import cli
from now.constants import Apps, DemoDatasets

os.environ['JCLOUD_LOGLEVEL'] = 'DEBUG'
app = Apps.TEXT_TO_IMAGE

params = [
    (Apps.TEXT_TO_IMAGE, DemoDatasets.BEST_ARTWORKS),
    (Apps.TEXT_TO_IMAGE, DemoDatasets.DEEP_FASHION),
    (Apps.IMAGE_TO_TEXT, DemoDatasets.RAP_LYRICS),
    (Apps.IMAGE_TO_IMAGE, DemoDatasets.TLL),
    (Apps.TEXT_TO_TEXT, DemoDatasets.ROCK_LYRICS),
    (Apps.MUSIC_TO_MUSIC, DemoDatasets.MUSIC_GENRES_MIX),
]


def deploy(app, data):
    kwargs = {
        'now': 'start',
        'app': app,
        'data': data,
        # 'custom_dataset_type': DatasetTypes.S3_BUCKET,
        # 'dataset_path': os.environ.get('S3_IMAGE_TEST_DATA_PATH'),
        # 'aws_access_key_id': os.environ.get('AWS_ACCESS_KEY_ID'),
        # 'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY'),
        # 'aws_region_name': 'eu-west-1',
        'deployment_type': 'remote',
        'proceed': True,
        'secured': False,
    }
    kwargs = Namespace(**kwargs)
    response = cli(args=kwargs)
    return response


if __name__ == '__main__':
    parallel = False
    results = []
    if parallel:
        with ProcessPoolExecutor() as thread_executor:
            futures = []
            for app, data in params:
                f = thread_executor.submit(deploy, app, data)
                futures.append(f)
            for f in futures:
                results.append(f.result())
    else:
        for app, data in params:
            result = deploy(app, data)
            results.append(result)
            print(result)
    print(results)



from docarray import Document, DocumentArray

from now.data_loading.convert_datasets_to_jpeg import to_thumbnail_jpg
from now.now_dataclasses import UserInput


def preprocess_images(da: DocumentArray) -> DocumentArray:
    """Loads all documents into memory to thumbnail them."""

    def convert_fn(d: Document):
        try:
            if d.tensor is None:
                if d.blob != b'':
                    d.convert_blob_to_image_tensor()
                elif d.uri:
                    d.load_uri_to_image_tensor()
            return to_thumbnail_jpg(d)
        except:
            return d

    for d in da:
        convert_fn(d)
    return DocumentArray(d for d in da if d.blob != b'')


def preprocess_text(da: DocumentArray, split_by_sentences=False) -> DocumentArray:
    """If necessary, loads text for all documents. If asked for, splits documents by sentences."""
    import nltk

    nltk.download('punkt', quiet=True)
    from nltk.tokenize import sent_tokenize

    def convert_fn(d: Document):
        try:
            if not d.text:
                if d.uri:
                    d.load_uri_to_text()
                    d.tags['additional_info'] = d.uri
            return d
        except:
            return d

    def gen_split_by_sentences():
        def _get_sentence_docs(batch):
            ret = []
            for d in batch:
                try:
                    ret += [
                        Document(
                            mime_type='text',
                            text=sentence,
                            tags=d.tags,
                        )
                        for sentence in set(sent_tokenize(d.text.replace('\n', ' ')))
                    ]
                except:
                    pass
            return ret

        for batch in da.map_batch(_get_sentence_docs, backend='process', batch_size=64):
            for d in batch:
                yield d

    for d in da:
        convert_fn(d)

    if split_by_sentences:
        da = DocumentArray(d for d in gen_split_by_sentences())

    return DocumentArray(d for d in da if d.text and d.text != '')


def preprocess_nested_docs(da: DocumentArray, user_input: UserInput) -> DocumentArray:
    """
    Process a `DocumentArray` with `Document`s that have `chunks` of nested `Document`s.
    It constructs `Document`s containg two chunks: one containing image data and another
    containing text data. Fields for indexing should be specified in the `UserInput`.

    :param da: A `DocumentArray` containing nested chunks.
    :param user_input: The configured user input.
    :return: A `DocumentArray` with `Document`s containing text and image chunks.
    """
    fields = user_input.task_config.indexer_scope
    texts, uris = [], []
    for doc in da:
        for chunk in doc.chunks:
            if chunk.tags['field_name'] == fields['text']:
                texts.append(chunk.content)
            elif chunk.tags['field_name'] == fields['image']:
                uris.append(chunk.content)
    return DocumentArray(
        [
            Document(
                chunks=[
                    Document(text=text[0]),
                    Document(uri=uri[0]),
                ]
            )
            for text, uri in zip(texts, uris)
        ]
    )



import json
import os
from typing import Dict, Optional, List

from docarray import DocumentArray, Document

from now.finetuning.run_finetuning import finetune
from now.finetuning.settings import parse_finetune_settings, FinetuneSettings
from now_common.preprocess import preprocess_nested_docs, preprocess_text

from now.apps.base.app import JinaNOWApp
from now.constants import Apps, Modalities, NOW_PREPROCESSOR_VERSION, ModelNames
from now.finetuning.data_builder import DataBuilder
from now.now_dataclasses import UserInput, DialogOptions, Task


class TextToTextAndImage(JinaNOWApp):
    """
    Hybrid text to text+image search combining symbolic and neural IR approaches.
    """

    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> Apps:
        return Apps.TEXT_TO_TEXT_AND_IMAGE

    @property
    def is_enabled(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return (
            'Text to text+image search app combining symbolic and neural IR approaches.'
        )

    def set_flow_yaml(self, **kwargs):
        """configure the flow yaml in the Jina NOW app."""
        flow_dir = os.path.abspath(os.path.join(__file__, '..'))
        self.flow_yaml = os.path.join(flow_dir, 'flow.yml')

    @property
    def input_modality(self) -> Modalities:
        return Modalities.TEXT

    @property
    def output_modality(self) -> Modalities:
        return Modalities.TEXT_AND_IMAGE

    @staticmethod
    def _create_task_config(user_input: UserInput, data_example: Document) -> Task:
        """
        Read task configuration template and replace field names.
        """
        # read task config template
        template_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), 'task_config.json'
        )
        with open(template_path) as f:
            config_dict = json.load(f)
            config = Task(**config_dict)
        # get text and image field names if they're not specified
        if not user_input.es_text_fields:
            user_input.es_text_fields = [
                chunk.tags['field_name']
                for chunk in data_example.chunks
                if chunk.modality == 'text'
            ]
        if not user_input.es_image_fields:
            user_input.es_image_fields = [
                chunk.tags['field_name']
                for chunk in data_example.chunks
                if chunk.modality == 'image'
            ]
        # put field names into generation function configurations
        for encoder in config.encoders:
            if encoder.name == 'text_encoder':
                for method in encoder.training_data_generation_methods:
                    method.query.scope = user_input.es_text_fields
                    method.target.scope = user_input.es_text_fields
            elif encoder.name == 'vision_encoder':
                for method in encoder.training_data_generation_methods:
                    method.query.scope = user_input.es_text_fields
                    method.target.scope = user_input.es_image_fields
        # specify text and image field for the indexer
        config.indexer_scope['text'] = user_input.es_text_fields[0]
        config.indexer_scope['image'] = user_input.es_image_fields[0]

        user_input.task_config = config
        return user_input.task_config

    def preprocess(
        self,
        da: DocumentArray,
        user_input: UserInput,
        is_indexing: Optional[bool] = False,
    ) -> DocumentArray:
        # Indexing
        if is_indexing:
            return preprocess_nested_docs(da=da, user_input=user_input)
        # Query
        else:
            return preprocess_text(da=da, split_by_sentences=False)

    def setup(
        self, dataset: DocumentArray, user_input: UserInput, kubectl_path
    ) -> Dict:
        task_config = self._create_task_config(user_input, dataset[0])
        data = DataBuilder(dataset=dataset, config=task_config).build()
        env_dict = {}
        for encoder_data, encoder_type in data:
            finetune_settings = self._construct_finetune_settings(
                user_input=user_input,
                dataset=dataset,
                encoder_type=encoder_type,
            )
            artifact_id, token = finetune(
                finetune_settings=finetune_settings,
                app_instance=self,
                dataset=encoder_data,
                user_input=user_input,
                env_dict={},
                kubectl_path=kubectl_path,
            )

            env_dict['JINA_TOKEN'] = token
            if finetune_settings.model_name == ModelNames.CLIP:
                env_dict['CLIP_ARTIFACT'] = artifact_id
            elif finetune_settings.model_name == ModelNames.SBERT:
                env_dict['SBERT_ARTIFACT'] = artifact_id
            else:
                print(f'{self.app_name} only expects CLIP or SBERT models.')
                raise

        env_dict[
            'PREPROCESSOR_NAME'
        ] = f'jinahub+docker://NOWPreprocessor/v{NOW_PREPROCESSOR_VERSION}'
        env_dict['APP'] = self.app_name
        self.set_flow_yaml()

        return env_dict

    def _construct_finetune_settings(
        self,
        user_input: UserInput,
        dataset: DocumentArray,
        encoder_type: str,
    ) -> FinetuneSettings:
        finetune_settings = parse_finetune_settings(
            user_input=user_input,
            dataset=dataset,
            model_name=self._model_name(encoder_type),
            loss=self._loss_function(encoder_type),
            add_embeddings=False,
        )
        # temporary adjustments to work with small text+image dataset
        finetune_settings.epochs = 2
        finetune_settings.num_val_queries = 5
        finetune_settings.train_val_split_ration = 0.8
        return finetune_settings

    @staticmethod
    def _model_name(encoder_type: Optional[str] = None) -> str:
        """Name of the model used in case of fine-tuning."""
        if encoder_type == 'text-to-text':
            return ModelNames.SBERT
        elif encoder_type == 'text-to-image':
            return ModelNames.CLIP

    @staticmethod
    def _loss_function(encoder_type: Optional[str] = None) -> str:
        """Loss function used during fine-tuning."""
        if encoder_type == 'text-to-image':
            return 'CLIPLoss'
        return 'TripletMarginLoss'



"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.
"""
from __future__ import annotations, print_function, unicode_literals

import importlib
import os
from typing import Dict, List

from hubble import AuthenticationRequiredError
from kubernetes import client, config

from now.constants import AVAILABLE_DATASET, Apps, DatasetTypes
from now.deployment.deployment import cmd
from now.log import time_profiler, yaspin_extended
from now.now_dataclasses import DialogOptions, UserInput
from now.thirdparty.PyInquirer import Separator
from now.utils import get_info_hubble, jina_auth_login, sigmap, to_camel_case

NEW_CLUSTER = {'name': '🐣 create new', 'value': 'new'}
AVAILABLE_SOON = 'will be available in upcoming versions'

# Make sure you add this dialog option to your app in order of dependency, i.e., if some dialog option depends on other
# then the parent should be called first before the dependant can called.

APP = DialogOptions(
    name='app',
    choices=[
        {
            'name': '📝 ▶ 🏞 text to image search',
            'value': Apps.TEXT_TO_IMAGE,
        },
        {
            'name': '🏞 ▶ 📝 image to text search',
            'value': Apps.IMAGE_TO_TEXT,
        },
        {
            'name': '🏞 ▶ 🏞 image to image search',
            'value': Apps.IMAGE_TO_IMAGE,
        },
        {'name': '📝 ▶ 📝 text to text search', 'value': Apps.TEXT_TO_TEXT},
        {
            'name': '📝 ▶ 🎦 text to video search (gif only at the moment)',
            'value': Apps.TEXT_TO_VIDEO,
        },
        {
            'name': '🥁 ▶ 🥁 music to music search',
            'value': Apps.MUSIC_TO_MUSIC,
        },
        {
            'name': '📝 ▶ 📝+🏞 text to text+image search',
            'value': Apps.TEXT_TO_TEXT_AND_IMAGE,
            'disabled': AVAILABLE_SOON,
        },
    ],
    prompt_message='What sort of search engine would you like to build?',
    prompt_type='list',
    is_terminal_command=True,
    description='What sort of search engine would you like to build?',
)

DATA = DialogOptions(
    name='data',
    prompt_message='What dataset do you want to use?',
    choices=lambda user_input, **kwargs: _get_data_choices(user_input, **kwargs),
    prompt_type='list',
    is_terminal_command=True,
    description='Select one of the available datasets or provide local filepath, '
    'docarray url, or docarray secret to use your own dataset',
    post_func=lambda user_input, **kwargs: _parse_custom_data_from_cli(
        user_input, **kwargs
    ),
)

CUSTOM_DATASET_TYPE = DialogOptions(
    name='custom_dataset_type',
    prompt_message='How do you want to provide input? (format: https://docarray.jina.ai/)',
    choices=[
        {
            'name': 'DocArray name (recommended)',
            'value': DatasetTypes.DOCARRAY,
        },
        {
            'name': 'DocArray URL',
            'value': DatasetTypes.URL,
        },
        {
            'name': 'Local path',
            'value': DatasetTypes.PATH,
        },
        {
            'name': 'S3 bucket',
            'value': DatasetTypes.S3_BUCKET,
        },
        {
            'name': 'Elasticsearch',
            'value': DatasetTypes.ELASTICSEARCH,
            'disabled': AVAILABLE_SOON,
        },
    ],
    prompt_type='list',
    depends_on=DATA,
    conditional_check=lambda user_input: user_input.data == 'custom',
)

DATASET_NAME = DialogOptions(
    name='dataset_name',
    prompt_message='Please enter your DocArray name:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.DOCARRAY,
)

DATASET_URL = DialogOptions(
    name='dataset_url',
    prompt_message='Please paste in the URL to download your DocArray from:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.URL,
)

DATASET_PATH = DialogOptions(
    name='dataset_path',
    prompt_message='Please enter the path to the local folder:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.PATH,
)

DATASET_PATH_S3 = DialogOptions(
    name='dataset_path',
    prompt_message='Please enter the S3 URI to the folder:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.S3_BUCKET,
)

AWS_ACCESS_KEY_ID = DialogOptions(
    name='aws_access_key_id',
    prompt_message='Please enter the AWS access key ID:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.S3_BUCKET,
)

AWS_SECRET_ACCESS_KEY = DialogOptions(
    name='aws_secret_access_key',
    prompt_message='Please enter the AWS secret access key:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.S3_BUCKET,
)

AWS_REGION_NAME = DialogOptions(
    name='aws_region_name',
    prompt_message='Please enter the AWS region:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.S3_BUCKET,
)

# --------------------------------------------- #

ES_TEXT_FIELDS = DialogOptions(
    name='es_text_fields',
    prompt_message='Please enter comma-separated text fields of your data:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.ELASTICSEARCH,
    post_func=lambda user_input, **kwargs: _parse_text_fields(user_input),
)


def _parse_text_fields(user_input: UserInput):
    user_input.es_text_fields = [
        field.strip() for field in user_input.es_text_fields.split(',')
    ]


ES_IMAGE_FIELDS = DialogOptions(
    name='es_image_fields',
    prompt_message='Please enter comma-separated image fields of your data:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input, **kwargs: user_input.custom_dataset_type
    == DatasetTypes.ELASTICSEARCH,
    post_func=lambda user_input, **kwargs: _parse_image_fields(user_input),
)


def _parse_image_fields(user_input: UserInput):
    user_input.es_image_fields = [
        field.strip() for field in user_input.es_image_fields.split(',')
    ]


ES_INDEX_NAME = DialogOptions(
    name='es_index_name',
    prompt_message='Please enter the name of your Elasticsearch index:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.ELASTICSEARCH,
)

ES_HOST_NAME = DialogOptions(
    name='es_host_name',
    prompt_message='Please enter the address of your Elasticsearch node:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.ELASTICSEARCH,
)

ES_ADDITIONAL_ARGS = DialogOptions(
    name='es_additional_args',
    prompt_message='Please enter additional arguments for your Elasticsearch node if there are any:',
    prompt_type='input',
    depends_on=CUSTOM_DATASET_TYPE,
    conditional_check=lambda user_input: user_input.custom_dataset_type
    == DatasetTypes.ELASTICSEARCH,
)

# --------------------------------------------- #

DEPLOYMENT_TYPE = DialogOptions(
    name='deployment_type',
    prompt_message='Where do you want to deploy your search engine?',
    prompt_type='list',
    choices=[
        {
            'name': '⛅️ Jina Cloud',
            'value': 'remote',
        },
        {
            'name': '📍 Local',
            'value': 'local',
        },
    ],
    is_terminal_command=True,
    description='Option is `local` and `remote`. Select `local` if you want search engine to be deployed on local '
    'cluster. Select `remote` to deploy it on Jina Cloud',
    post_func=lambda user_input, **kwargs: _jina_auth_login(user_input, **kwargs),
)

LOCAL_CLUSTER = DialogOptions(
    name='cluster',
    prompt_message='On which cluster do you want to deploy your search engine?',
    prompt_type='list',
    choices=lambda user_input, **kwargs: _construct_local_cluster_choices(
        user_input, **kwargs
    ),
    depends_on=DEPLOYMENT_TYPE,
    is_terminal_command=True,
    description='Reference an existing `local` cluster or select `new` to create a new one. '
    'Use this only when the `--deployment-type=local`',
    conditional_check=lambda user_inp: user_inp.deployment_type == 'local',
    post_func=lambda user_input, **kwargs: _check_requirements(user_input, **kwargs),
)

PROCEED = DialogOptions(
    name='proceed',
    prompt_message='jina-now is deployed already. Do you want to remove the current data?',
    prompt_type='list',
    choices=[
        {'name': '⛔ no', 'value': False},
        {'name': '✅ yes', 'value': True},
    ],
    depends_on=LOCAL_CLUSTER,
    conditional_check=lambda user_input: _check_if_namespace_exist(),
)

SECURED = DialogOptions(
    name='secured',
    prompt_message='Do you want to secure the flow?',
    prompt_type='list',
    choices=[
        {'name': '✅ yes', 'value': True},
        {'name': '⛔ no', 'value': False},
    ],
    depends_on=DEPLOYMENT_TYPE,
    conditional_check=lambda user_inp: user_inp.deployment_type == 'remote',
)

ADDITIONAL_USERS = DialogOptions(
    name='additional_user',
    prompt_message='Do you want to provide additional users access to this flow?',
    prompt_type='list',
    choices=[
        {'name': '✅ yes', 'value': True},
        {'name': '⛔ no', 'value': False},
    ],
    depends_on=SECURED,
    conditional_check=lambda user_inp: user_inp.secured,
)

USER_EMAILS = DialogOptions(
    name='user_emails',
    prompt_message='Please enter the comma separated Email IDs '
    'who will have access to this flow:',
    prompt_type='input',
    depends_on=ADDITIONAL_USERS,
    conditional_check=lambda user_inp: user_inp.additional_user,
    post_func=lambda user_input, **kwargs: _add_additional_users(user_input, **kwargs),
)


def _add_additional_users(user_input: UserInput, **kwargs):
    user_input.user_emails = (
        [email.strip() for email in kwargs['user_emails'].split(',')]
        if kwargs['user_emails']
        else []
    )


def _check_if_namespace_exist():
    config.load_kube_config()
    v1 = client.CoreV1Api()
    return 'nowapi' in [item.metadata.name for item in v1.list_namespace().items]


def _construct_app(app_name: str):
    return getattr(
        importlib.import_module(f'now.apps.{app_name}.app'),
        f'{to_camel_case(app_name)}',
    )()


@time_profiler
def _check_requirements(user_input: UserInput, **kwargs) -> None:
    user_input.app_instance.run_checks(user_input)


def _jina_auth_login(user_input, **kwargs):
    if user_input.deployment_type != 'remote':
        return

    try:
        jina_auth_login()
    except AuthenticationRequiredError:
        with yaspin_extended(
            sigmap=sigmap, text='Log in to JCloud', color='green'
        ) as spinner:
            cmd('jina auth login')
        spinner.ok('🛠️')

    get_info_hubble(user_input)
    os.environ['JCLOUD_NO_SURVEY'] = '1'


def _get_data_choices(user_input, **kwargs) -> List[Dict[str, str]]:
    app_instance = user_input.app_instance
    return [
        {'name': name, 'value': value}
        for value, name in AVAILABLE_DATASET[app_instance.output_modality]
    ] + [
        Separator(),
        {
            'name': '✨ custom',
            'value': 'custom',
        },
    ]


def _construct_local_cluster_choices(user_input, **kwargs):
    active_context = kwargs.get('active_context')
    contexts = kwargs.get('contexts')
    context_names = _get_context_names(contexts, active_context)
    choices = [NEW_CLUSTER]
    # filter contexts with `gke`
    if len(context_names) > 0 and len(context_names[0]) > 0:
        context_names = [
            context
            for context in context_names
            if 'gke' not in context and _cluster_running(context)
        ]
        choices = context_names + choices
    return choices


def _get_context_names(contexts, active_context=None):
    names = [c for c in contexts] if contexts is not None else []
    if active_context is not None:
        names.remove(active_context)
        names = [active_context] + names
    return names


def _cluster_running(cluster):
    config.load_kube_config(context=cluster)
    v1 = client.CoreV1Api()
    try:
        v1.list_namespace()
    except Exception as e:
        return False
    return True


def _parse_custom_data_from_cli(user_input: UserInput, **kwargs) -> None:
    data = user_input.data
    if data == 'custom':
        return

    app_instance = user_input.app_instance
    for k, v in enumerate(AVAILABLE_DATASET[app_instance.output_modality]):
        if v[0] == data:
            return

    try:
        data = os.path.expanduser(data)
    except Exception:
        pass
    if os.path.exists(data):
        user_input.custom_dataset_type = DatasetTypes.PATH
        user_input.dataset_path = data
    elif 'http' in data:
        user_input.custom_dataset_type = DatasetTypes.URL
        user_input.dataset_url = data
    else:
        user_input.custom_dataset_type = DatasetTypes.DOCARRAY
        user_input.dataset_name = data


data = [DATA]
data_da = [CUSTOM_DATASET_TYPE, DATASET_NAME, DATASET_PATH, DATASET_URL]
data_s3 = [DATASET_PATH_S3, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, AWS_REGION_NAME]
data_es = [
    ES_HOST_NAME,
    ES_INDEX_NAME,
    ES_TEXT_FIELDS,
    ES_IMAGE_FIELDS,
    ES_ADDITIONAL_ARGS,
]
cluster = [DEPLOYMENT_TYPE, LOCAL_CLUSTER]
remote_cluster = [SECURED, ADDITIONAL_USERS, USER_EMAILS]


base_options = data + data_da + data_s3 + data_es + cluster + remote_cluster



from now.utils import BetterEnum

# TODO: Uncomment the DEMO_DATASET_DOCARRAY_VERSION when the DocArray datasets on GCloud has been changed
# from docarray import __version__ as docarray_version
# DEMO_DATASET_DOCARRAY_VERSION = docarray_version
DEMO_DATASET_DOCARRAY_VERSION = '0.13.17'

DOCKER_BFF_PLAYGROUND_TAG = '0.0.127-fix-music-to-music-app-4'
NOW_PREPROCESSOR_VERSION = '0.0.88-fix-music-to-music-app-4'
NOW_ANNLITE_INDEXER_VERSION = '0.0.6-annlite-update-list-endpoint-4'


class Modalities(BetterEnum):
    TEXT = 'text'
    IMAGE = 'image'
    MUSIC = 'music'
    VIDEO = 'video'
    TEXT_AND_IMAGE = 'text_and_image'


class Apps(BetterEnum):
    TEXT_TO_TEXT = 'text_to_text'
    TEXT_TO_IMAGE = 'text_to_image'
    IMAGE_TO_TEXT = 'image_to_text'
    IMAGE_TO_IMAGE = 'image_to_image'
    MUSIC_TO_MUSIC = 'music_to_music'
    TEXT_TO_VIDEO = 'text_to_video'
    TEXT_TO_TEXT_AND_IMAGE = 'text_to_text_and_image'


class DatasetTypes(BetterEnum):
    DEMO = 'demo'
    PATH = 'path'
    URL = 'url'
    DOCARRAY = 'docarray'
    S3_BUCKET = 's3_bucket'
    ELASTICSEARCH = 'elasticsearch'


class Qualities(BetterEnum):
    MEDIUM = 'medium'
    GOOD = 'good'
    EXCELLENT = 'excellent'


class ModelNames(BetterEnum):
    MLP = 'mlp'
    SBERT = 'sentence-transformers/msmarco-distilbert-base-v3'
    CLIP = 'openai/clip-vit-base-patch32'


BASE_STORAGE_URL = (
    'https://storage.googleapis.com/jina-fashion-data/data/one-line/datasets'
)

CLIP_USES = {
    'local': ('CLIPOnnxEncoder/latest', 'ViT-B-32::openai', 512),
    'remote': ('CLIPOnnxEncoder/latest-gpu', 'ViT-B-32::openai', 512),
}

EXTERNAL_CLIP_HOST = 'encoderclip-bh-5f4efaff13.wolf.jina.ai'

PREFETCH_NR = 10


class DemoDatasets:
    BEST_ARTWORKS = 'best-artworks'
    NFT_MONKEY = 'nft-monkey'
    TLL = 'tll'
    BIRD_SPECIES = 'bird-species'
    STANFORD_CARS = 'stanford-cars'
    DEEP_FASHION = 'deepfashion'
    NIH_CHEST_XRAYS = 'nih-chest-xrays'
    GEOLOCATION_GEOGUESSR = 'geolocation-geoguessr'
    MUSIC_GENRES_ROCK = 'music-genres-mid'
    MUSIC_GENRES_MIX = 'music-genres-mix'
    ROCK_LYRICS = 'rock-lyrics'
    POP_LYRICS = 'pop-lyrics'
    RAP_LYRICS = 'rap-lyrics'
    INDIE_LYRICS = 'indie-lyrics'
    METAL_LYRICS = 'metal-lyrics'
    TUMBLR_GIFS = 'tumblr-gifs'
    TUMBLR_GIFS_10K = 'tumblr-gifs-10k'
    ES_ONLINE_SHOP_50 = 'extracted-data-online-shop-50-flat'


AVAILABLE_DATASET = {
    Modalities.IMAGE: [
        (DemoDatasets.BEST_ARTWORKS, '🖼 artworks (≈8K docs)'),
        (DemoDatasets.NFT_MONKEY, '💰 nft - bored apes (10K docs)'),
        (DemoDatasets.TLL, '👬 totally looks like (≈12K docs)'),
        (DemoDatasets.BIRD_SPECIES, '🦆 birds (≈12K docs)'),
        (DemoDatasets.STANFORD_CARS, '🚗 cars (≈16K docs)'),
        (DemoDatasets.GEOLOCATION_GEOGUESSR, '🏞 geolocation (≈50K docs)'),
        (DemoDatasets.DEEP_FASHION, '👕 fashion (≈53K docs)'),
        (DemoDatasets.NIH_CHEST_XRAYS, '☢️ chest x-ray (≈100K docs)'),
    ],
    Modalities.MUSIC: [
        (DemoDatasets.MUSIC_GENRES_ROCK, '🎸 rock music (≈2K songs)'),
        (DemoDatasets.MUSIC_GENRES_MIX, '🎸 multiple genres (≈2K songs)'),
    ],
    Modalities.TEXT: [
        (DemoDatasets.ROCK_LYRICS, '🎤 rock lyrics (200K docs)'),
        (DemoDatasets.POP_LYRICS, '🎤 pop lyrics (200K docs)'),
        (DemoDatasets.RAP_LYRICS, '🎤 rap lyrics (200K docs)'),
        (DemoDatasets.INDIE_LYRICS, '🎤 indie lyrics (200K docs)'),
        (DemoDatasets.METAL_LYRICS, '🎤 metal lyrics (200K docs)'),
    ],
    Modalities.VIDEO: [
        (DemoDatasets.TUMBLR_GIFS, '🎦 tumblr gifs (100K gifs)'),
        (DemoDatasets.TUMBLR_GIFS_10K, '🎦 tumblr gifs subset (10K gifs)'),
    ],
    Modalities.TEXT_AND_IMAGE: [
        (DemoDatasets.ES_ONLINE_SHOP_50, 'online shop data (50 products)')
    ],
}

SURVEY_LINK = 'https://10sw1tcpld4.typeform.com/to/VTAyYRpR?utm_source=cli'
